/** \file statisticsGaps.cpp
 * \brief \ref statisticsGaps method implementation file.
 *
 * In this file the \ref statisticsGaps class methods are implemented.
 */

#include "statisticsGaps.h"

#define DEPURACION2
		
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  statisticsGaps::statisticsGaps(char **, int, int)	|
|       Class constructor.                              |
|							|
+++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

statisticsGaps::statisticsGaps(char **alignmentMatrix, int species, int aminos) {
  
  int i, j;
  
  columnLength = species;
  columns = 	 aminos;
  maxGaps = 	 0;
  halfWindow  =  0;
  
  /* Memory allocation for the vectors, and initialization */
  gapsInColumn = 	new int[columns];
  utils::initlVect(gapsInColumn, columns, 0);
  
  aminosXInColumn = 	new int[columns];
  utils::initlVect(aminosXInColumn, aminos, 0);	
  
  gapsWindow = 		new int[columns];
  utils::initlVect(gapsWindow, columns, 0);
  
  numColumnsWithGaps = 	new int[species];
  utils::initlVect(numColumnsWithGaps, columnLength, 0);
  
  /* Count the gaps and indeterminations of each columns */
  for(i = 0; i < columns; i++) {
    for(j = 0; j < columnLength; j++){
      switch(alignmentMatrix[j][i]) {
      case '-':
	  gapsInColumn[i]++; break;
      case 'X':
	  aminosXInColumn[i]++; break;
      }      
    }
    
    /* Increase the number of colums with the number of gaps of the last processed column */
    numColumnsWithGaps[gapsInColumn[i]]++;
    gapsWindow[i] = gapsInColumn[i];
    if(gapsWindow[i] > maxGaps) maxGaps = gapsWindow[i];
  }

}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  statisticsGaps::statisticsGaps()	                |
|       Null class constructor.                         |
|							|
+++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

statisticsGaps::statisticsGaps(){
  
  /* Initializate all values to NULL or 0 */
  gapsInColumn = NULL;
  numColumnsWithGaps = NULL;
  aminosXInColumn = NULL;
  gapsWindow = NULL;
  
  columns = 	  0;
  columnLength  = 0;	
  maxGaps = 	  0;
  halfWindow = 	  0;

}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  statisticsGaps::~statisticsGaps()	                |
|       Class destructor                                |
|							|
++++++++++++++++ +++++++++++++++++++++++++++++++++++++++*/

statisticsGaps::~statisticsGaps() {

  /* Only free memory if there is previous memory allocation */
  if(gapsInColumn != NULL){
    delete[] gapsInColumn;
    delete[] numColumnsWithGaps;
    delete[] aminosXInColumn;
    delete[] gapsWindow;
  }
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  bool statisticsGaps::applyWindow(int)		  |
|	Method that calculates the number of gaps of each |
|	column using the average of its value and 	  |
|	the half window near values in the right and      |
|       in the left.                                      |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

bool statisticsGaps::applyWindow(int mediumWinSize) {
  
  int i, j;

  /* If one of this conditions is true, we return FALSE:                         */
  /*    .- If already exists a previously calculated vector for this window size */
  /*    .- If mediumWinSize is negative                                          */
  /*    .- If mediumWinSize value is greater than 1/4 of alignment length        */
  
  if(mediumWinSize < 0 || mediumWinSize == halfWindow || mediumWinSize > columns/4)
     return false;
  
  /* Initializate to 0 the vector that will store the number of gaps of each column */
  /* and the vector that will store window processing results                       */
  utils::initlVect(numColumnsWithGaps, columnLength, 0);
  utils::initlVect(gapsWindow, columns, 0);
  
  /* Initializate max number of gaps per column value and store the mediumWinSize value */
  /* in the object                                                                      */
  maxGaps = 0;
  halfWindow = mediumWinSize;
  int window = (2 * mediumWinSize + 1);
  
  /* We calculate some statistics for every column in the alignment, */
  /* and the maximum number of gaps value                            */
  for(i = 0; i < columns; i++) {
    /* Sum the total number of gaps for the considered window */
    for(j = i - mediumWinSize; j <= i + mediumWinSize; j++) {
      if(j < 0) 		gapsWindow[i] += gapsInColumn[-j];
      else if(j >= columns) 	gapsWindow[i] += gapsInColumn[2*columns - j - 2];
      else			gapsWindow[i] += gapsInColumn[j];
    }
    
    /* Calculate, and round to the nearest integer, the number of gaps for the i column */
    gapsWindow[i] = utils::roundInt(((double) gapsWindow[i]/window));

    /* Increase in 1 the number of colums with the same number of gaps than column i */
    numColumnsWithGaps[gapsWindow[i]]++;
    
    /* Update the max number of gaps in the alignment, if neccesary */
    if(gapsWindow[i] > maxGaps) maxGaps = gapsWindow[i];
  }

  return true;
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  int statisticsGaps::calcCutPoint(float, float)		|
|	Method that calculates the cut point that the alignment |
|	class will use to clean the alignment.                  |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int statisticsGaps::calcCutPoint(float gapBaseLine, float gapThreshold) {
  
  int i, cutThr, cutBL, acum;
  
  /* Calculate the number of gaps representing the % of gaps, fixed by gapThreshold.          */
  /* This number of gaps will be the maximum number of gaps permitted in the clean alignment. */
  /* Round to the nearest integer.                                                            */

  cutThr = utils::roundInt(((double) (columnLength * gapThreshold) / 100.0));
  
  /* Now, calculate the number of gaps representing the % of the columns, fixed by gapBaseLine */
  /* This number of gaps is the cut value that ensures us that we conserve that % of columns   */
  cutBL = utils::roundInt(((double) (columns * gapBaseLine) / 100.0));

  /* We look the number of gaps that permit us to conserve at least the % of columns given by gapBaseLine */
  for(i = 0, acum = 0; i < columnLength; i++) {
    
    acum += numColumnsWithGaps[i];
    
    if(acum >= cutBL) {
      cutBL = i; break;
    }
    
  }

  /* Return the maximum value of the two calculated cuts (gapThreshold cut and gapBaseLine cut) */
  return (utils::max(cutThr, cutBL));
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  int statisticsGaps::calcCutPointMixSlope()			|
|	Method that calculates, automaticly, the cut point that |
|	the alignment class will use to clean the alignment.    |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int statisticsGaps::calcCutPointMixSlope(void) {
  
  int prev, pprev, maxIter, row = 1, act = 1, max = 0;
  float *firstSlopeVector, *secondSlopeVector, delta = 0, maxSlope = -1;

  /* We build two slope vectors, one vector for the first slope and another one for the second. */
  firstSlopeVector = new float[maxGaps+1];
  secondSlopeVector = new float[maxGaps+1];
  utils::initlVect(firstSlopeVector, maxGaps, -1.0);
  utils::initlVect(secondSlopeVector, maxGaps, -1.0);		 
  maxIter = maxGaps + 1;

  while(act < maxIter) {

    /* We look for a first point to second slope. */
    while((numColumnsWithGaps[act]) == 0) act++; 
    pprev = act;

    /* We look for a first point to first slope. */
    do { act++; } while((numColumnsWithGaps[act]) == 0);
    prev = act;  

    /* We look for a second point to first and second slope. */
    do { act++; } while((numColumnsWithGaps[act]) == 0);

    if(act >= maxIter) break;

    /* Calculate the first slope between the earlier previus and previus points. */
    firstSlopeVector[prev] =  ((float) (prev - pprev) / columnLength);
    firstSlopeVector[prev] /= ((float) numColumnsWithGaps[prev] / columns);

    /* Calculate the first slope between the previus and current points. */
    firstSlopeVector[act] =  ((float) (act - prev) / columnLength);
    firstSlopeVector[act] /= ((float) numColumnsWithGaps[act] / columns);

    /* Calculate the second slope between the earlier previus and current points. */
    secondSlopeVector[act] = ((float) (act - pprev) / columnLength);
    secondSlopeVector[act] /= ((float) (numColumnsWithGaps[act] + numColumnsWithGaps[prev]) / columns);

    /* If the ratio between ... */       
    if((secondSlopeVector[pprev] != -1.0) || (firstSlopeVector[pprev] != -1.0)) {

      /* .- first slope previus and first slope earlier previus points. */
      if(firstSlopeVector[pprev] != -1.0) {
        delta = firstSlopeVector[prev]/firstSlopeVector[pprev];
        row = pprev;
      }

      /* .- first slope current and first slope previus points. */
      if(delta < (firstSlopeVector[act]/firstSlopeVector[prev])) {
        delta = firstSlopeVector[act]/firstSlopeVector[prev];
        row = prev;
      }

      /* .- second slope current and second slope earlier previus points. */
      if(secondSlopeVector[pprev] != -1.0) {
        if(delta < (secondSlopeVector[act]/secondSlopeVector[pprev])) { 
          delta = secondSlopeVector[act]/secondSlopeVector[pprev];
          row = pprev;
        }
      }

      /* ... is better that current maxSlope then we modify the maxSlope with the best ratio. */
      if(delta > maxSlope) { 
        maxSlope = delta;
        max = row;
      }
    }
  }
  delete[] firstSlopeVector;
  delete[] secondSlopeVector;
  return max;  
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  int statisticsGaps::calcCutPointMixSlope()			|
|	Method that calculates, automaticly, the cut point that |
|	the alignment class will use to clean the alignment.    |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int statisticsGaps::calcCutPoint2ndSlope(void) {
  
  int prev, pprev, maxIter, act = 1, max = 0;
  float *secondSlopeVector, delta = 0, maxSlope = -1;

  /* We build two slope vectors, one vector for the first slope and another one for the second. */
  secondSlopeVector = new float[maxGaps+1];
  utils::initlVect(secondSlopeVector, maxGaps, -1.0);		 
  maxIter = maxGaps + 1;

  while(act < maxIter) {

    /* We look for a first point to second slope. */
    while((numColumnsWithGaps[act]) == 0) act++; 
    pprev = act;

    /* We look for a first point to first slope. */
    do { act++; } while((numColumnsWithGaps[act]) == 0);
    prev = act;

    /* We look for a second point to first and second slope. */
    do { act++; } while((numColumnsWithGaps[act]) == 0);

    if(act >= maxIter) break;

    /* Calculate the second slope between the earlier previus and current points. */
    secondSlopeVector[act] = ((float) (act - pprev) / columnLength);
    secondSlopeVector[act] /= ((float) (numColumnsWithGaps[act] + numColumnsWithGaps[prev]) / columns);

    /* If the ratio between second slope current and second slope earlier previus points. */
    if(secondSlopeVector[pprev] != -1.0) {
      if(delta < (secondSlopeVector[act]/secondSlopeVector[pprev]))
        delta = secondSlopeVector[act]/secondSlopeVector[pprev];

      /* ... is better that current maxSlope then we modify the maxSlope with the new value. */
      if(delta > maxSlope) { 
        maxSlope = delta;
        max = pprev;
      }
    }
  }

  delete[] secondSlopeVector;
  return max;  
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  int *statisticsGaps::getGapsWindow(void)		        |
|       This method returns a pointer to the                    |
|       gapsWindow vector, used then for threating and cleaning |
|       the alingment.                                          | 
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int *statisticsGaps::getGapsWindow(void) {
  
  return gapsWindow;
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  int *statisticsGaps::getIndVector(void)		              |
|       This method returns a pointer to the                          |
|       indeterminations vector, used then for conservation value     | 
|       calculation (class statisticsConservation)                    |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int *statisticsGaps::getIndVector(){

  return aminosXInColumn;
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  int *statisticsGaps::getHalfWindow(void)		              |
|       This method returns the half window value of the              |
|       statisticsGaps object                                         |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int statisticsGaps::getHalfWindow(){

  return halfWindow;
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  void statisticsGaps::printGapsColumns(void)		|
|	This method prints the gap statistics of each   |						|
|	column.					        |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void statisticsGaps::printGapsColumns(void) {
  
  cout.precision(10);
  cout << "Column\t%Gaps" << endl;

  for(int i = 0; i < columns; i++)
    cout << setiosflags(ios::left) << setw(4) << i << "\t" << setw(7) << (gapsWindow[i] * 100.0)/columnLength  << endl;
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++
|  void statisticsGaps::printGapsAcl(void)		|
|	This method prints the accumulated statistics   | 
|       related to the number of                        |
|       columns for a determinated number of gaps       |
+++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void statisticsGaps::printGapsAcl(void) {
  int acm, i;  

  cout << "N.Col\t%Length\t\tN.Gaps\t%Gaps" << endl;
  cout.precision(10);

  for(i = 0, acm = 0; i <= maxGaps; i++) {
    if(numColumnsWithGaps[i] != 0) { 
      acm += numColumnsWithGaps[i];
      cout << setiosflags(ios::left) << acm << "\t" << setw(10) << (acm * 100.0)/columns << "\t" << i 
           << "\t" << setw(10) << (i * 100.0)/columnLength << "\t" << endl;
    }
  } 
}

